/*

 ideserv.c - An incomplete but mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>
 Win32 port by
  Josef Soucek <josef.soucek@ide64.org>
 Serial support by
  Borsanyi Attile <bigfoot@axelero.hu>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"
#ifndef WIN32
#include <sys/io.h>
#include <argp.h>
#else
#include <stdio.h>
#include <errno.h>
#include "resource.h"
#endif
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>
#include <sys/time.h>
#include <pwd.h>
#include <grp.h>
#include <err.h>
#include <string.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/resource.h>
#include <stdlib.h>
#include <time.h>
#ifdef RS232
#include <termios.h>
#endif

FILE *lf=NULL;
unsigned char buffer[16][65536];
unsigned short int buffbytes[16];
unsigned short int buffp[16];

#include "ideserv.h"
#include "crc8.h"

#define CM_CLOSED 0
#define CM_FILE 1
#define CM_DIR 2
#define CM_ERR 3

#define OPEN_RONLY 0
#define OPEN_STATUS 2
#define OPEN_WONLY 4
#define OPEN_ERR 6

#ifdef WIN32
/* Prototypes */
INT_PTR CALLBACK MyDlgProc(HWND,UINT,WPARAM,LPARAM);
long WINAPI PCLinkLoop(long);
#endif

const char *argp_program_version="ideservd 0.19";
FILE *files[16];
int channelmode[16];

void convertfilename(char *s, char typek, char *vessz, char *name, char *type, unsigned char *mode) {
    int b,i,j;

    i=j=*vessz=*mode=0;

    for (b=0;s[b] && s[b]!='/';b++) {
        if (s[b]==':') {i=*vessz=j=*mode=0; continue;}

        switch (*vessz) {
        case 0:
            if (s[b]==typek) {
                (*vessz)++;
                break;
            }
            if (i<16) name[i++]=s[b];
            break;
        case 1:
            if (s[b]==',') {
                (*vessz)++;
                break;
            }
            if (j<3) type[j++]=s[b];
            break;
        case 2:if (mode) *mode=s[b];(*vessz)++;break;
        }
    }
    name[i]=0;
    type[j]=0;
    if (j==1) switch (type[0]) {
        case 'S':strcpy(type,"SEQ");break;
        case 'P':strcpy(type,"PRG");break;
        case 'U':strcpy(type,"USR");break;
        case 'B':strcpy(type,"DIR");break;
        case 'J':strcpy(type,"LNK");break;
    }
}

int resolv_path(char *s, char *keszpath, int resolv_all) {
    unsigned char mode;
    char lname[1000],name[17],type[4],vessz;
    struct dirent *ep;
    DIR *dp;
    int found;
    int fel=0,fel2=0;

    while (s[fel] && s[fel]!=':') fel++;
    fel=s[fel]?1:0;

    if (fel) {
        while (s[fel2] && s[fel2]!='/' && s[fel2]!=':') s[fel2++]='/';
        fel=fel2+1;
    }

    if (s[fel]=='/') {
        keszpath[0]='/';
        keszpath[1]=0;
        found=1;
    } else {
        keszpath[0]=0;
        found=0;
    }
    fel=0;

    while (1) {

        while (s[fel] && (s[fel]=='/' || s[fel]==':')) fel++;

        if (!s[fel]) {
            if (!resolv_all) s[0]=0;
            return found;
        }

        if (!resolv_all) {
            fel2=fel;
            while (s[fel2] && s[fel2]!='/' && s[fel2]!=':') fel2++;
            if (!s[fel2]) {
                fel2=fel;fel=0;
                while (s[fel2]) s[fel++]=s[fel2++];
                s[fel]=0;
                return 1;
            }
        }

        dp=opendir(keszpath[0]?keszpath:".");
        if (!dp) {errtochannel15(1);fprintf(lf,"Couldn't open the directory \"%s\".\n",keszpath);return 0;}
        convertfilename(s+fel,'/',&vessz,name,type,&mode);

        while ((ep=readdir(dp)))
        {
            struct stat buf;
            DIRECTORY_ENTRY_T dirent;
            char lname2[2020];

            strcpy(lname2,keszpath);
            strncpy(lname,ep->d_name,999);
            strncat(lname2,ep->d_name,999);

            if (lstat(lname2,&buf)) continue;

            createdirentry(ep,&buf,&dirent);

            if ((dirent.attrib & (A_DIR | A_CLOSED))!=(A_CLOSED | A_DIR)) continue;

            dirent.filesize[0]=0;
            if (!matchname(dirent.name,name)) continue;

            found=1;
            break;
        }
        closedir(dp);

        if (!found)
        {
            int f=0,g=fel;
            while (s[g]) {
                lname[f]=c64toascii[(unsigned char)s[g]];
                f++;g++;
            }
            lname[f]=0;
        }

        if ((strlen(lname)+strlen(keszpath)+2)>1000) return 0;

        strcat(keszpath,lname);
        strcat(keszpath,"/");
        while (s[fel] && s[fel]!='/' && s[fel]!=':') fel++;

        if (!found) return 0;
    }
}

void do_chdir(char *s) {
    char keszpath[1000];

    if (!resolv_path(s,keszpath,1))
    {
        if (!s[0]) seterror(34,0);
        else seterror(39,0);
        fprintf(lf,"No such directory called \"%s\".\n",keszpath);
        return;
    }
    fprintf(lf,"Chdir to \"%s\".\n",keszpath);
    errtochannel15(chdir(keszpath));
}

void do_mkdir(char *s) {
    unsigned char mode;
    struct dirent *ep;
    DIR *dp;
    char keszpath[1020],vessz,type[4],lname[1000];
    int found=0;

    while (s[found] && s[found]!=':') found++;
    if (!s[found]) {seterror(30,0);found=0;keszpath[0]=0;goto vege2;}

    found=0;
    if (!resolv_path(s,keszpath,0)) {seterror(39,0);goto vege2;}

    convertfilename(s,',',&vessz,s,type,&mode);

    dp=opendir(keszpath[0]?keszpath:".");
    if (!dp) {errtochannel15(1);fprintf(lf,"Couldn't open the directory \"%s\".\n",keszpath);return;}

    while ((ep=readdir(dp)))
    {
        struct stat buf;
        DIRECTORY_ENTRY_T dirent;
        char lname2[2020];

        strcpy(lname2,keszpath);
        strncpy(lname,ep->d_name,999);
        strncat(lname2,ep->d_name,999);

        if (lstat(lname2,&buf)) continue;

        createdirentry(ep,&buf,&dirent);

        if ((dirent.attrib & A_ANY)!=A_DIR) continue;

        dirent.filesize[0]=0;
        if (!matchname(dirent.name,s)) continue;

        found=1;
        break;
    }

    closedir(dp);

    if (!found) {
        int f=0;
        while (s[f]) {
            if (f<16) {
                lname[f]=c64toascii[(unsigned char)s[f]];
                f++;
            }
        }
        lname[f]=0;
    }

    strcat(keszpath,lname);

    if (found)
        seterror(63,0);
    else {
        int f=0;

        while (lname[f]) {
            if (lname[f]=='*' || lname[f]=='?' || lname[f]==':' || lname[f]=='=') {
                seterror(33,0);goto vege2;
            }
            f++;
        }
        errtochannel15(mkdir(keszpath,0777));
    }
vege2:
    if (!s[0]) {
        seterror(34,0);
    }
    fprintf(lf,"Mkdir \"%s\".\n",keszpath);
}

void do_scratch(char *s, int smode) {
    unsigned char mode;
    char lname[1000],type[4],vessz,keszpath[1000];
    struct dirent *ep;
    DIR *dp;
    int found=0;

    while (s[found] && s[found]!=':') found++;
    if (!s[found]) {seterror(30,0);found=0;keszpath[0]=0;goto vege;}

    found=0;
    if (!resolv_path(s,keszpath,0)) {seterror(39,0);goto vege;}

    convertfilename(s,'=',&vessz,s,type,&mode);

    dp=opendir(keszpath[0]?keszpath:".");
    if (!dp) {errtochannel15(1);fprintf(lf,"Couldn't open the directory \"%s\".\n",keszpath);return;}
    if (type[0]==0) strcpy(type,"*");

    while ((ep=readdir(dp)))
    {
        struct stat buf;
        DIRECTORY_ENTRY_T dirent;
        char lname2[2020];

        strcpy(lname2,keszpath);
        strncpy(lname,ep->d_name,999);
        strncat(lname2,ep->d_name,999);

        if (lstat(lname2,&buf)) continue;

        createdirentry(ep,&buf,&dirent);

        if (smode) {
            if ((dirent.attrib & (A_ANY | A_CLOSED))!=(A_DIR | A_CLOSED)) continue;
        }
        else
            if ((dirent.attrib & A_ANY)==A_DIR) continue;

        if (!(dirent.attrib & A_DELETEABLE)) continue;

        dirent.filesize[0]=0;
        if (!matchname(dirent.name,s)) continue;
        if (!smode) {
            dirent.time[0]=0;
            if (!matchname(dirent.filetype,type)) continue;
        }

        found=1;
        break;
    }

    closedir(dp);
    if (!found) seterror(1,0);
vege:
    if (!found) {
        int f=0,g=0;
        while (s[f]) {
            lname[f]=c64toascii[(unsigned char)s[f]];
            f++;
        }
        if (!smode) {
            lname[f++]=',';
            while (type[g]) {
                lname[f]=c64toascii[(unsigned char)type[g]];
                f++;g++;
            }
        }
        lname[f]=0;
    }

    strcat(keszpath,lname);

    if (found) {
        if (smode)
            errtochannel15(rmdir(keszpath));
        else
            errtochannel15(unlink(keszpath));
    }

    if (!s[0]) {
        seterror(34,0);
    }
    if (smode)
        fprintf(lf,"Rmdir \"%s\"\n",keszpath);
    else
        fprintf(lf,"Rm \"%s\"\n",keszpath);
}

void commandchannel(char *s) {
    if (!strncmp(s,"CD:",3) || !strncmp(s,"CD/",3)) {
        s[2]=':';
        do_chdir(s+2);
    } else if (s[0]=='M' && s[1]=='D') {
        do_mkdir(s);
    } else if (s[0]=='R' && s[1]=='D') {
        do_scratch(s,1);
    } else if (s[0]=='S' && s[1]!='-') {
        do_scratch(s,0);
    } else if (!strncmp(s,"CD_",3)) {
        errtochannel15(chdir(".."));
        fprintf(lf,"Chdir \"..\"\n");
    } else if (!strncmp(s,"C\xd0",2)) {
        seterror(2,1);
        fprintf(lf,"Change to partition %d\n",s[3]);
    } else if (!strncmp(s,"UI",2)) {
        seterror(73,0);
        fprintf(lf,"Identify device\n");
    } else if (!strncmp(s,"I",1)) {
        seterror(0,0);
        fprintf(lf,"Initialize\n");
    } else if (!strncmp(s,"V",1)) {
        seterror(0,0);
        fprintf(lf,"Validate\n");
    } else if (s[0]) {
        seterror(31,0);
    } else fprintf(lf,"\n");
}

void openfile(int usecrc) {
    char s[256];
    unsigned char status=OPEN_ERR;
    unsigned char channel,ch,fp;
    unsigned long long l=0;

    ch=getb();if (tout) return;
    channel=ch & 0x0f;

    fp=0;
    while (fp<255) {
        s[fp]=getb();if (tout) return;
        if (!s[fp]) break;
        fp++;
    }
    s[fp]=0;

    if (usecrc) {
        unsigned char crc=crc8[0xCF],gcrc,gt;
        crc=crc8[crc ^ ch];
        fp=0;
        while (s[fp]) { crc=crc8[crc ^ (unsigned char)s[fp]]; fp++;}
	crc=crc8[crc];
        gcrc=getb();if (tout) return;
        crc=crc8[crc ^ gcrc];
        gt=getb();if (tout) return;
        if (crc) {
	    seterror(23,1);
            fprintf(lf,"CRC error %02X: CF %02X ",crc,ch);
            fp=0;
            while (s[fp]) {fprintf(lf,"%02X ",(unsigned char)s[fp]);fp++;}
            fprintf(lf,"00 %02X %02X\n",gcrc,gt);
            return;
        }
    }
#ifndef RS232
    alarm(0);
#endif

    if (channel==15) {
	channelmode[15]=CM_ERR;
        fprintf(lf,"Open [%d]: \"%s\": ",channel,s);
	commandchannel(s);
	status=OPEN_STATUS;
    } else {
        char vessz,type[4];
        unsigned char mode;
        
        if (channelmode[channel]==CM_FILE) fclose(files[channel]);
        channelmode[channel]=CM_CLOSED;

        if (s[0]=='$') {
            unsigned char dirline[32];
            struct dirent *ep;
            DIR *dp;
            int sum=0,partition;
            char keszpath[1000];
            channelmode[channel]=CM_DIR;
            buffbytes[channel]=buffp[channel]=0;

	    partition=!strncmp(&s[1],"=P",2);

            resolv_path(s+1,keszpath,0);
            fprintf(lf,"Open [%d]: \"$%s%s\": ",channel,keszpath,s+1);
            convertfilename(s+1,'=',&vessz,s,type,&mode);

            if (s[0]==0) strcpy(s,"*");
            if (type[0]==0) strcpy(type,"*");

            if (channel==1) {
                status=OPEN_ERR;
            } else
            if (channel>1) {
                dp=opendir(keszpath[0]?keszpath:".");
                if (!dp) {sendb(2);errtochannel15(1);fprintf(lf,"Couldn't open the directory \"%s\".\n",keszpath);return;}
                buffer[channel][buffbytes[channel]++]='I';

                memset(&dirline[0],32,16);
                memset(&dirline[16],0,8);
                dirline[24]=A_DIR;
                memcpy(&dirline[25],"DIR",3);
                memset(&dirline[28],0,4);
                for (l=0;l<32;l++) buffer[channel][buffbytes[channel]++]=dirline[l];

                while ((ep=readdir(dp)))
                {
                    struct stat buf;
                    char lname2[2020];

                    strcpy(lname2,keszpath);
                    strncat(lname2,ep->d_name,999);

                    if (lstat(lname2,&buf)) continue;

                    createdirentry(ep,&buf,(DIRECTORY_ENTRY_T*)dirline);

                    for (l=0;l<32;l++) buffer[channel][buffbytes[channel]++]=dirline[l];
                }

                errtochannel15(closedir(dp));
                status=OPEN_RONLY;
            } else

            if (partition) {
                memcpy(&dirline[0x00],"\1\4\1\1\377\0\22\42",8);
                memset(&dirline[0x08],32,16);
                memcpy(&dirline[0x18],"\42\40LINUX\0",8);
                for (l=0;l<32;l++) buffer[channel][buffbytes[channel]++]=dirline[l];
                dirline[0x00]=dirline[0x01]=1;
                dirline[0x02]=1;
                dirline[0x03]=0;
                memcpy(&dirline[0x04],"   \42PARTITION 1\42      NET  \0",28);
                for (l=0;l<32;l++) buffer[channel][buffbytes[channel]++]=dirline[l];
                dirline[0x00]=dirline[0x01]=1;
                dirline[0x02]=1;
                dirline[0x03]=0;
                memcpy(&dirline[0x04],"PARTITIONS.              \0\0\0",28);
                for (l=0;l<32;l++) buffer[channel][buffbytes[channel]++]=dirline[l];
                errtochannel15(0);
                status=OPEN_RONLY;

            } else {

                dp=opendir(keszpath[0]?keszpath:".");
                if (!dp) {sendb(2);errtochannel15(1);fprintf(lf,"Couldn't open the directory \"%s\".\n",keszpath);return;}
//                sendb(OPEN_RONLY);
                memcpy(&dirline[0x00],"\1\4\1\1\1\0\22\42",8);
                memset(&dirline[0x08],32,16);
#ifdef WIN32
                memcpy(&dirline[0x18],"\42\40WIN32\0",8);
#else
                memcpy(&dirline[0x18],"\42\40LINUX\0",8);
#endif
                for (l=0;l<32;l++) buffer[channel][buffbytes[channel]++]=dirline[l];

                while ((ep=readdir(dp)))
                {
                    struct stat buf;
                    DIRECTORY_ENTRY_T dirent;
                    int offs;
                    unsigned int n;
                    char lname2[2020];

                    strcpy(lname2,keszpath);
                    strncat(lname2,ep->d_name,999);

                    if (lstat(lname2,&buf)) continue;

                    createdirentry(ep,&buf,&dirent);

                    if (dirent.pointer[0] & 0x80) continue;
                    n=(dirent.filesize[0]+(dirent.filesize[1] << 8)+(dirent.filesize[2] << 16)+(dirent.filesize[3] << 24)+255) >> 8;
                    if (n>65535) n=65535;

                    dirent.filesize[0]=0;
                    if (!matchname(dirent.name,s)) continue;
                    dirent.time[0]=0;
                    if (!matchname(dirent.filetype,type)) continue;

                    dirline[0x00]=dirline[0x01]=1;
                    dirline[0x02]=n;
                    dirline[0x03]=n >> 8;
                    sum+=n;
                    memset(&dirline[0x04],32,27);
                    offs=7;
                    if (n>9) offs--;
                    if (n>99) offs--;
                    if (n>999) offs--;
                    dirline[offs++]=0x22;
                    n=0;while (dirent.name[n] && n<16) dirline[offs++]=dirent.name[n++];
                    dirline[offs++]=0x22;
                    offs+=(16-n);

                    dirline[offs++]=(dirent.attrib & A_CLOSED)?' ':'*';

                    n=0;while (n<3 && dirent.filetype[n]) {dirline[offs+n]=dirent.filetype[n];n++;}
                    if ((dirent.attrib & (A_DELETEABLE | A_WRITEABLE))!=(A_DELETEABLE | A_WRITEABLE)) dirline[offs+3]='<';

                    for (l=0;l<32;l++) buffer[channel][buffbytes[channel]++]=dirline[l];
                    if (sum>65535) sum=65535;
                }

                dirline[0x00]=dirline[0x01]=1;
                dirline[0x02]=sum;
                dirline[0x03]=sum >> 8;
                memcpy(&dirline[0x04],"BLOCKS USED.             \0\0\0",28);

                for (l=0;l<32;l++) buffer[channel][buffbytes[channel]++]=dirline[l];
                errtochannel15(closedir(dp));
                status=OPEN_RONLY;
            }
        } else {
            char vessz,type[4];
            unsigned char mode,overwrite;
            char lname[1000],keszpath[1020];
            struct dirent *ep;
            DIR *dp;
            int found=0;
            int b;

            overwrite=0;
            for (b=0;s[b];b++) if (s[b]==':') {
                if (s[0]=='@') overwrite=1;
                break;
            }

            resolv_path(s,keszpath,0);
            convertfilename(s,',',&vessz,s,type,&mode);

            dp=opendir(keszpath[0]?keszpath:".");
            if (!dp) {sendb(2);errtochannel15(1);fprintf(lf,"Couldn't open the directory \"%s\".\n",keszpath);return;}
            if (type[0]==0) strcpy(type,(channel==1)?"PRG":"*");

            while ((ep=readdir(dp)))
            {
                struct stat buf;
                DIRECTORY_ENTRY_T dirent;
                char lname2[2020];

                strcpy(lname2,keszpath);
                strncpy(lname,ep->d_name,999);
                strncat(lname2,ep->d_name,999);

                if (lstat(lname2,&buf)) continue;

                createdirentry(ep,&buf,&dirent);

                if ((dirent.attrib & (A_ANY | A_CLOSED))!=(A_CLOSED | A_NORMAL)) continue;

                dirent.filesize[0]=0;
                if (!matchname(dirent.name,s)) continue;
                dirent.time[0]=0;
                if (!matchname(dirent.filetype,type)) continue;

                found=1;
                break;
            }

            closedir(dp);
            status=OPEN_ERR;

            if (!found) {
                int f=0,g=0;
                while (s[f]) {
                    if (f<16) {
                        lname[f]=c64toascii[(unsigned char)s[f]];
                        f++;
                    }
                }
                lname[f++]=',';
                while (type[g]) {
                    if (g<3) {
                        lname[f]=c64toascii[(unsigned char)type[g]];
                        f++;g++;
                    }
                }
                lname[f]=0;
            }

            if (channel==0) mode='R';
            if (channel==1) mode='W';

            strcat(keszpath,lname);

            fprintf(lf,"Open [%d]: \"%s\": ",channel,keszpath);

            if (!s[0]) {
                seterror(34,0);goto vege;
            }

            if (mode=='W') {
                int f=0;

                while (lname[f]) {
                    if (lname[f]=='*' || lname[f]=='?' || lname[f]==':' || lname[f]=='=') {
                        seterror(33,0);goto vege;
                    }
                    f++;
                }

                if (found && !overwrite) {
                    seterror(63,0);
                } else {
                    if (overwrite) f=open(keszpath, O_CREAT|O_WRONLY|O_TRUNC, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
                    else f=open(keszpath, O_CREAT|O_EXCL|O_WRONLY, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);

                    if (f<0) errtochannel15(1); else {
                        status=OPEN_WONLY;
                        files[channel]=fdopen(f,"wb");
                        channelmode[channel]=CM_FILE;
                        errtochannel15(0);
                    }
                }
            } else {
                if (!found) {
                    seterror(62,0);
                } else {
                    files[channel]=fopen(keszpath,"rb");
                    if (files[channel]) {
                        status=OPEN_RONLY;
                        channelmode[channel]=CM_FILE;
                        errtochannel15(0);
                    } else errtochannel15(1);
                }
            }
        }
    }
vege:
    sendb(status);
}

void closefile(int usecrc) {
    unsigned char channel,ch;
    channel=(ch=getb()) & 0x0f;if (tout) return;
    if (usecrc) {
        unsigned char crc=crc8[0xC3];
        crc=crc8[crc ^ ch];
        crc=crc8[crc ^ getb()];if (tout) return;
	getb();if (tout) return;
        if (crc) {seterror(23,1);fprintf(lf,"CRC error\n");return;}
    } else {
	getb();if (tout) return;
    }
    sendb(0);
    fprintf(lf,"Close [%d]\n",channel);
    if (channelmode[channel]==CM_FILE) fclose(files[channel]);
    channelmode[channel]=CM_CLOSED;
}

void readfile(int usecrc) {
    unsigned char channel,ch,status=0;
    unsigned short int bytes,i,j;
    struct timeval start,end;
    unsigned long long l=0;

    channel=(ch=getb()) & 0x0f;if (tout) return;
    bytes=getb();if (tout) return;
    bytes|=getb()*256;if (tout) return;
    if (usecrc) {
        unsigned char crc=crc8[0xD2];
        crc=crc8[crc ^ ch];
        crc=crc8[crc ^ (bytes & 0xff)];
        crc=crc8[crc ^ ((bytes >> 8) & 0xff)];
        crc=crc8[crc ^ getb()];if (tout) return;
	getb();if (tout) return;
        if (crc) {seterror(23,1);fprintf(lf,"CRC error\n");return;}
    } else {
	getb();if (tout) return;
    }
    fprintf(lf,"Read [%d] %d bytes\n",channel,bytes);
    sendb(0);if (tout) return;
    if (channelmode[channel]==CM_CLOSED) {
	sendb(0);if (tout) return;
	sendb(0);if (tout) return;
	if (usecrc) {
	    unsigned char crc=crc8[crc8[0]];
	    sendb(crc);if (tout) return;
	}
	sendb(0);return;
    }
    if (channelmode[channel]==CM_ERR) {
	j=buffp[channel];
	if (bytes==(buffbytes[channel]-j)) status=0x40;
	if (bytes>buffbytes[channel]-j) {bytes=buffbytes[channel]-j;status=0x42;}
	l=bytes;
	buffp[channel]+=bytes;
    } else {
	if (channelmode[channel]==CM_FILE) {
	    unsigned long itt;
	    l=bytes=fread(buffer[channel],1,bytes?bytes:65536,files[channel]);
	    if (feof(files[channel])) status=0x42; else
	    {
		itt=ftell(files[channel]);
		fgetc(files[channel]);
		if (feof(files[channel])) status=0x40;
		fseek(files[channel],itt,SEEK_SET);
	    }
	    j=0;
	} else {
	    j=buffp[channel];
	    if (bytes==(buffbytes[channel]-j)) status=0x40;
	    if (bytes>buffbytes[channel]-j) {bytes=buffbytes[channel]-j;status=0x42;}
	    l=bytes;
	    buffp[channel]+=bytes;
	}
    }
    sendb(~bytes);if (tout) return;
    sendb(~bytes >> 8);if (tout) return;
#ifdef RS232
    alarm(3);
#ifndef WIN32
    tcdrain(commhandle);
#else
    FlushFileBuffers(commhandle);
#endif
    alarm(0);if (tout) return;
#endif
    gettimeofday(&start,NULL);
    if (bytes) {
	unsigned short int j2=j;
	for (i=0;i<bytes;i++,j++) {sendb(buffer[channel][j]);if (tout) return;}
	if (usecrc) {
	    unsigned char crc=0;
	    crc=crc8[crc ^ (unsigned char)(~bytes)];
	    crc=crc8[crc ^ (unsigned char)(~bytes >> 8)];
	    for (i=0;i<bytes;i++,j2++) {crc=crc8[crc ^ buffer[channel][j2]];}
//    fprintf(lf,"crc: %02X\n", crc);
	    sendb(crc);if (tout) return;
	}
    }else {
	unsigned short int j2=j;
	for (i=0;i<32768;i++,j++) {sendb(buffer[channel][j]);if (tout) return;}
	for (i=0;i<32768;i++,j++) {sendb(buffer[channel][j]);if (tout) return;}
	if (usecrc) {
	    unsigned char crc=0;
	    for (i=0;i<32768;i++,j2++) {crc=crc8[crc ^ buffer[channel][j2]];}
	    for (i=0;i<32768;i++,j2++) {crc=crc8[crc ^ buffer[channel][j2]];}
	    sendb(crc);if (tout) return;
	}
    }
    gettimeofday(&end,NULL);
#ifndef WIN32 
// Temporarily removed in WIN32 Floating point exeption (core dumped) SIGFE Arithmetic exception
    l=l*1000000/((end.tv_sec-start.tv_sec)*1000000+end.tv_usec-start.tv_usec);
    fprintf(lf,"Read %d bytes at %lld bytes/seconds\n",bytes,l);
#endif
    sendb(status);//fprintf(lf,"Status: %02X\n",status);
#ifdef RS232
    alarm(1);
#ifndef WIN32
    tcdrain(commhandle);
#else
    FlushFileBuffers(commhandle);
#endif
    alarm(0);if (tout) return;
#endif
    if (channel==15 && buffbytes[15]==buffp[15]) seterror(0,0);
}

void writefile(int usecrc) {
    unsigned char channel,ch;
    unsigned short int bytes,i;
    struct timeval start,end;
    unsigned long long l=0;

    channel=(ch=getb()) & 0x0f;if (tout) return;
    bytes=getb();if (tout) return;
    bytes|=getb()*256;if (tout) return;
    if (usecrc) {
        unsigned char crc=crc8[0xD7];
        crc=crc8[crc ^ ch];
        crc=crc8[crc ^ (bytes & 0xff)];
        crc=crc8[crc ^ ((bytes >> 8) & 0xff)];
        crc=crc8[crc ^ getb()];if (tout) return;
	getb();if (tout) return;
        if (crc) {sendb(2);seterror(23,1);fprintf(lf,"CRC error\n");return;}
    } else {
	getb();if (tout) return;
    }
    fprintf(lf,"Write [%d] %d bytes\n",channel,bytes);
    if (channelmode[channel]==CM_ERR) {
        char s[256];
        unsigned char j;
        sendb(0);if (tout) return;
        j=0;
        for (i=0;i<bytes;i++) {s[j]=getb();if (tout) return;if (j<255) j++;}

	if (usecrc) {
    	    unsigned char crc=0;

	    for (i=0;i<j;i++) crc=crc8[crc ^ (unsigned char)s[i]];
    	    crc=crc8[crc ^ getb()];if (tout) return;
	    getb();if (tout) return;
    	    if (crc) {sendb(2);seterror(23,1);fprintf(lf,"CRC error\n");return;}
	} else {
    	    getb();if (tout) return;
	}

        s[j]=0;
        if (j && s[j-1]==0x0d) s[j-1]=0;

        sendb(0);
        commandchannel(s);
    } else {
        if (channelmode[channel]!=CM_FILE) {fprintf(lf,"Not writeable.\n");sendb(2);return;}
        sendb(0);if (tout) return;

        gettimeofday(&start,NULL);
        for (i=0;i<bytes;i++) {buffer[channel][i]=getb();if (tout) return;}
        gettimeofday(&end,NULL);
	
	if (usecrc) {
    	    unsigned char crc=0;

	    for (i=0;i<bytes;i++) crc=crc8[crc ^ buffer[channel][i]];
    	    crc=crc8[crc ^ getb()];if (tout) return;
	    getb();if (tout) return;
    	    if (crc) {sendb(2);seterror(23,1);fprintf(lf,"CRC error\n");return;}
	} else {
    	    getb();if (tout) return;
	}
	
        if (fwrite(buffer[channel],1,bytes,files[channel])==bytes) sendb(0); else sendb(2);
        l=bytes;
#ifndef WIN32 
// Temporarily removed in WIN32 Floating point exeption (core dumped) SIGFE Arithmetic exception
        l=l*1000000/((end.tv_sec-start.tv_sec)*1000000+end.tv_usec-start.tv_usec);
        fprintf(lf,"Wrote %d bytes at %lld bytes/seconds\n",bytes,l);
#endif

    }
}

const char *argp_program_bug_address="<soci@c64.rulez.org>";
static char doc[]="IDEDOS 0.9x PCLink fileserver";

struct arguments
{
    int background;
    int priority;
    char *user;
    char *group;
    char *root;
    char *log;
#ifdef PARPORT
    char *parport;
#endif
#ifdef RS232
    char *serport;
#endif
    int hog;
};

#ifndef WIN32
static struct argp_option options[]={
    {"user",		'u',	"USER",		0,  "User under we run (nobody)"},
    {"group",		'g',	"GROUP",	0,  "Group under we run (nogroup)"},
    {"root",		'r',	"DIRECTORY",	0,  "Root directory (.)"},
    {"log",		'l',	"FILE",		0,  "Logfile (stdout)"},
    {"nice",		'n',	"ADJUST",	0,  "Adjust nice level (19)"},
#ifndef RS232
    {"hog",		'h',	0,		0,  "CPU hog (off)"},
#endif
    {"background",	'b',	0, 		0,  "Fork into background"},
#ifdef PARPORT
    {"device",		'd',	"PARPORT",	0,  "Parport device"},
#endif
#ifdef RS232
    {"device",		'd',	"SERIALPORT",	0,  "Serial port device"},
#endif
    { 0 }
};

static error_t parse_opt (int key,char *arg,struct argp_state *state)
{
    struct arguments *arguments=state->input;

    switch (key)
    {
    case 'b':arguments->background=1;break;
    case 'h':arguments->hog=1;break;
    case 'u':arguments->user=arg;break;
    case 'g':arguments->group=arg;break;
    case 'l':arguments->log=arg;break;
    case 'r':arguments->root=arg;break;
#ifdef PARPORT
    case 'd':arguments->parport=arg;break;
#endif
#ifdef RS232
    case 'd':arguments->serport=arg;break;
#endif
    case 'n':arguments->priority=strtol(arg,NULL,0);break;
    case ARGP_KEY_ARG:argp_usage(state);break;
    case ARGP_KEY_END:break;
    default:return ARGP_ERR_UNKNOWN;
    }
    return 0;
}

static struct argp argp={options,parse_opt,NULL,doc};
#endif

int vtime=1;

void init(int argc,char *argv[])
{
    struct passwd *user2;
    struct group *group2;
    struct arguments arguments;
    arguments.background=0;
    arguments.user= "nobody";
    arguments.group= "nogroup";
    arguments.root= ".";
#ifdef WIN32
    arguments.log= "ideservd.log";
#else
    arguments.log= "/dev/stdout";
#endif
#ifdef PARPORT
    arguments.parport= "/dev/parport0";
#endif
#ifdef RS232
#ifdef WIN32
    arguments.serport= "COM1";
#else
    arguments.serport= "/dev/ttyS0";
#endif
#endif
    arguments.priority=19;
    arguments.hog=0;
#ifndef WIN32
    argp_parse(&argp, argc, argv, 0, 0, &arguments);
#endif

#ifdef RS232
    initport(arguments.serport);
#else
#ifdef WIN32
#ifndef WIN32_IO
     /* Load the library */
     hLib = LoadLibrary("inpout32.dll");

     if (hLib == NULL) {
          err(1,"LoadLibrary Failed.\n");
     }

     /* get the address of the function */

     inp32 = (inpfuncPtr) GetProcAddress(hLib, "Inp32");

     if (inp32 == NULL) {
          err(1,"GetProcAddress for Inp32 Failed.\n");
     }

     oup32 = (oupfuncPtr) GetProcAddress(hLib, "Out32");

     if (oup32 == NULL) {
          err(1,"GetProcAddress for Oup32 Failed.\n");
     }
#endif // !WIN32_IO
#else

#ifdef PARPORT
    parportfd=open(arguments.parport, O_RDWR);
    if (parportfd == -1) err(1,"Opening parport device");
    if (ioctl(parportfd, PPCLAIM)) {
	close(parportfd);
	err(1,"Cannot access parport device");
    }
    {
	int mode=IEEE1284_MODE_BYTE;
	if (ioctl(parportfd, PPSETMODE, &mode)) {
	    close(parportfd);
	    err(1,"Cannot set byte mode on parport device");
	}
	
	mode=0;
	if (ioctl(parportfd, PPDATADIR, &mode)) {
	    close(parportfd);
	    err(1,"Cannot set direction on parport device");
	}
    }
#else
    if (ioperm(IOPORT,3,1)) err(1,"Allocating IO ports");
#endif // PARPORT
#endif // WIN32 (LPT port)
#endif // RS232

#ifndef WIN32
    if (!(user2=getpwnam(arguments.user))) errx(1,"User %s does not exists.",arguments.user);
    if (!(group2=getgrnam(arguments.group))) errx(1,"Group %s does not exists.",arguments.group);
// chroot crashes on WIN32 with current cygwin1.dll without registry entry:
// [HKEY_LOCAL_MACHINE\SOFTWARE\Cygnus Solutions\Cygwin\mounts v2\/]
// "native"="C:\\"
#endif
    if (!(lf=fopen(arguments.log,"wt"))) err(1,"Opening logfile '%s'",arguments.log);
#ifndef WIN32
    if (chroot(arguments.root) || chdir("/")) err(1,"Changing root dir to '%s'",arguments.root);
#endif
#ifndef WIN32
    if (setgid(group2->gr_gid) || setuid(user2->pw_uid)) errx(1,"Could not drop privileges.");
    if (arguments.background) if (fork()) errx(0,"Forking into background.");
    nice(arguments.priority);
    if (arguments.hog) vtime=0;
    fclose(stdin);
    fclose(stderr);
    fclose(stdout);
#endif
}

#ifdef WIN32
INT_PTR CALLBACK MyDlgProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
  switch(message) 
  {
   case WM_INITDIALOG:
        return TRUE;

   case WM_COMMAND:
      switch(wParam) 
      {
         case IDC_EXIT:
	 PostQuitMessage(0);
         EndDialog(hwnd, TRUE);
         return TRUE;
      }
   break;
  }
return FALSE;
}
#endif

int main(int argc,char *argv[])
//int WINAPI WinMain (HINSTANCE hInstance,
//                    HINSTANCE hPrevInstance,
//                    LPSTR szCmdLine,
//                    int iCmdShow)
{
    unsigned char b;
    init(argc,argv);
    fprintf(lf,"%s\n",argp_program_version);
    signal(SIGALRM,timeout);
    signal(SIGTERM,terminate);
    signal(SIGINT,terminate);
    for (b=0;b<16;b++) channelmode[b]=CM_CLOSED;
    seterror(73,0);

#ifdef WIN32
    HWND hwnd;
    MSG msg;
    hwnd=CreateDialog(NULL, MAKEINTRESOURCE(IDD_MAIN), NULL, (DLGPROC)MyDlgProc);
    ShowWindow(hwnd, SW_SHOW);	// probably not necessary...
    HANDLE hThread;
    DWORD iID;
    hThread = CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)PCLinkLoop,NULL,0,&iID);
    
    while(GetMessage (&msg,NULL,0,0)!=0) {
//       TranslateMessage(&msg);  // Only if there is text input
       DispatchMessage(&msg);
    }
    TerminateThread(hThread,0);
    terminate();
}

long WINAPI PCLinkLoop(long lParam)
{
    unsigned char b;
#endif

start:
    fprintf(lf,"ready.\n");
start2:
#ifndef RS232
    alarm(0);
#ifdef PC64
#ifdef PARPORT
    for (;;) {
	unsigned char b;
	ioctl(parportfd, PPRSTATUS, &b);
	if (b>=0x80) {
	    ioctl(parportfd, PPRSTATUS, &b);
	    if (b & 0x78) break;
	}
	if (vtime) usleep(100000);
    }
#else
    do {
	while (inb(STATPORT)<0x80) if (vtime) usleep(100000);
    } while ((inb(STATPORT) & 0x78)==0);
#endif // PARPORT
    tout=0;alarm(1);
#else // PC64
    out_b(CLKLO+DATAHI+ATNHI);
    wdatalo() if (vtime) usleep(100000);
    out_b(CLKHI+DATAHI+ATNHI);
    tout=0;alarm(1);
    wdatahi() {if (vtime) usleep(10000);if (tout) goto start;} //set clklo, wait atnlo
#endif // PC64
#else // RS232
    tout=2;
#endif // RS232

    b=getb();
#ifdef RS232
    if (tout) goto start2;
#else
    if (tout) goto start;
    alarm(0);
#endif
    switch (b)
    {
//    case 'X':return 0;
    case 0x00:goto start2;
    case 'O':openfile(0);break;
    case 0xCF:openfile(1);break;
    case 'R':readfile(0);break;
    case 0xD2:readfile(1);break;
    case 'W':writefile(0);break;
    case 0xD7:writefile(1);break;
    case 'C':closefile(0);break;
    case 0xC3:closefile(1);break;
    default:fprintf(lf,"Unknown command %02X\n",b);
    }
#ifdef RS232
    alarm(1);tout=0;
#ifndef WIN32
    tcdrain(commhandle);
#else
    FlushFileBuffers(commhandle);
#endif
    alarm(0);
#endif
    goto start;
}
